#!/usr/bin/perl -w
#
# mtreehmm.pl - EM algorithm for mutagenetic tree hidden Markov models (MtreeHMMs)
#
# Version 0.1.01 Jul 2006
#
#
# Copyright (C) 2005-2006 Niko Beerenwinkel
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#


use strict;
use Getopt::Std;

use lib "$ENV{HOME}/perllib";   # modify this as necessary
use Math::Amoeba qw(MinimiseND);



# GLOBAL CONSTANTS:

my $EpsEps = 1e-10;     # box contraints [EpsEps, 1-EpsEps] for lambda, eps_p, and eps_n
my $InitialEps = 0.01;  # initial guess for eps_p and eps_n


# COMMAND LINE SWITCHES:

# defaults:
my %option = ( E => 0.0458552,  # for Bacheler data
               I => 100,
               A => 1e-3
             );

my $usage = "
MtreeHMM - Parameter estimation for the  mutagenetic tree hidden Markov model

[ Version 0.1.01 Jul 2006,  Copyright (C) 2005-2006 Niko Beerenwinkel ]

Usage:  
   mtreehmm.pl -f [-CEABIpnh]

Flags: 
   -f <filestem>     Filestem to use (required)
   -B <#samples>     Number of bootstrap samples    
   -C <time>         Constant sampling <time>, _or_
   -E <rate>         Exponential sampling time at <rate> (default = 0.0458552)
   -A <diff>         Accuracy of EM, min. log-likelihood difference (default = 1e-3)
   -I <iter>         Max. nunmber of EM iterations (default = 100)
   -p                Uniform false positive rate eps+
   -n                Uniform false negative rate eps-
   -h                This help

Input files:
   [filestem.prf]    List of mutation names (optional; (1, ..., n), if omitted)
   filestem.model    Mutagenetic tree model, adjacency matrix
   filestem.patient  List of patients
   filestem.week     List of timepoints
   filestem.pat      List of binary mutational patterns 

Output files:
   filestem.lambda   Poisson rates for mutation to occur
   filestem.epsp     False positives rates
   filestem.epsn     False negatives rates
   filestem.l        Likelihood
   filestem.viterbi  Inferred population state (MAP estimate)

See also:
   * http://math.berkeley.edu/~niko/mtreehmm
   * N. Beerenwinkel and M. Drton.
     A mutagenetic tree hidden Markov model for longitudinal clonal HIV sequence data. 
     Biostatistics, 2006, in press. http://arxiv.org/abs/q-bio/0603031
   * Mtreemix, http://mtreemix.bioinf.mpi-sb.mpg.de/

";

getopts("f:C:E:A:B:I:pnh", \%option);

if (scalar(@ARGV) >= 1 || $option{h}) {
  die $usage;
}
if (($option{C} && $option{E}) || ((! $option{C}) && (! $option{E}))) {
  die "Use either -C or -E!\n\n" . $usage;
}
if (! $option{f}) {
    die "Use -f to specify the filestem of your data files!\n\n" . $usage;
}

my $filestem = $option{f};
print "\nFilestem  = $filestem\n";
my ($sampling_mode, $sampling_param);
if ($option{C}) {
  $sampling_mode  = 'CONSTANT';
  $sampling_param = $option{C};
}
else {
  $sampling_mode  = 'EXPONENTIAL';
  $sampling_param = $option{E};
} 
my $accuracy = $option{A};
my $max_iter = $option{I};
my $B = 0;
if ($option{B}) {
  $B = $option{B};
}


# MAIN:

my @FIXEDPARAM;  # global array
my (@lambda, @mu, @offset, @eps_p, @eps_n);  # parameters
my @states;  # hidden states
my @loglike;

my ($allpatients, $timepoint, $clones) = load_data($filestem);

my ($events, $parent, $lambda_init, $compatibles, $is_pred_state);  # model
($events, $parent, $lambda_init, $compatibles, $is_pred_state) = load_model($filestem, $sampling_mode, $sampling_param);
print_model($parent, $lambda_init, $compatibles, $is_pred_state);


($loglike[0], $lambda[0], $mu[0], $offset[0], $eps_p[0], $eps_n[0], $states[0]) = ExpectationMaximization($parent, $lambda_init, $compatibles, $is_pred_state, $allpatients, $timepoint, $clones, $accuracy, $max_iter, $option{p}, $option{n}, 1);

print_viterbi($filestem, $allpatients, $timepoint, scalar(@{$parent}), $mu[0], $offset[0], $states[0], $clones); 

my %dummy;
print_param(0, $filestem, $allpatients, $events, $loglike[0], $lambda[0], $mu[0], \%dummy, $offset[0], \%dummy, $eps_p[0], $eps_n[0]);


for my $b (1 .. $B) {
  my ($allpatients_b, $timepoint_b, $clones_b) = resample($allpatients, $timepoint, $clones);

  my ($mu_b, %mu_b_list, $offset_b, %offset_b_list);
  ($loglike[$b], $lambda[$b], $mu_b, $offset_b, $eps_p[$b], $eps_n[$b], $states[$b]) = ExpectationMaximization($parent, $lambda_init, $compatibles, $is_pred_state, $allpatients_b, $timepoint_b, $clones_b, $accuracy, $max_iter, $option{p}, $option{n}, 0);

  for my $p (0 .. $#{$allpatients_b}) {
    push @{$mu_b_list{$allpatients_b->[$p]}}, $mu_b->[$p];
    push @{$offset_b_list{$allpatients_b->[$p]}}, $offset_b->[$p];
  }

  print_param($b, $filestem, $allpatients_b, $events, $loglike[$b], $lambda[$b], $mu_b, \%mu_b_list, $offset_b, \%offset_b_list, $eps_p[$b], $eps_n[$b]);
}
 

# END MAIN



sub ExpectationMaximization {

  my @parent        = @{shift @_};
  my @lambda        = @{shift @_};
  my @compatibles   = @{shift @_};
  my @is_pred_state = @{shift @_};
  my @allpatients   = @{shift @_};
  my @timepoint     = @{shift @_};
  my @clones        = @{shift @_};
  my $accuracy      = shift @_;
  my $max_iter      = shift @_;
  my $uniform_eps_p = shift @_;
  my $uniform_eps_n = shift @_;
  my $do_viterbi    = shift @_;

  my ($p, $t, $r, $s, $c, $u, $v, $a, $b, $delta_t);
  my (@theta0, %theta);
  my $loglike     = -Inf;    # log-likelihood of the data
  my $loglike_new = -1e100;  # new value


  # initial parameter guess:
  my ($eps_p, $eps_n) = guess_eps($#parent);
  my @mu = guess_mu(scalar(@allpatients));
  my @offset = guess_offset(scalar(@allpatients));


  # EM iteration:
  # =============

  my $iter = 1;
  while (($iter <= $max_iter) && (($loglike_new - $loglike > $accuracy) || ($iter == 1))) {
    print "\niter = $iter\n";
    $iter++;
    $loglike = $loglike_new;
    $loglike_new = 0.0;

    my @flips = ();  # expected no of. flips
    for $v (1 .. $#parent) {
      $flips[$v][0][0] = $flips[$v][0][1] = 0.0;  # pseudo count
      $flips[$v][1][0] = $flips[$v][1][1] = 0.0;  # pseudo count
    }

    my @trans0 = ();   # expected initial states
    my @trans  = ();   # expected no. of transitions

    for $p (0 .. $#clones) {  # for all patients
      printf "%3d %3d  (%s)\n", $iter-1, $p, $allpatients[$p];

      # E step:
      # -------

      # precompute transition matrices:
      @theta0 = initial_state_loglike(\@compatibles, \@parent, \@lambda, $mu[$p], $offset[$p], $timepoint[$p]);
      %theta = all_transition_matrices(\@timepoint, \@parent, \@lambda, $mu[$p], $offset[$p], $compatibles, \@is_pred_state);  # indexed by time difference delta_t

      my @forward = forward_pass(\@theta0, \%theta, \@compatibles, $timepoint[$p], $eps_p, $eps_n, $clones[$p]);

      my @backward = backward_pass(\@theta0, \%theta, \@compatibles, $timepoint[$p], $eps_p, $eps_n, $clones[$p]);
 
      # compute log-likelihood of the data
      my $N_t = scalar(@{$timepoint[$p]});
      my $fsum = $forward[$N_t-1][0];
      for $s (1 .. $#compatibles) {
        $fsum = logsum($fsum, $forward[$N_t-1][$s]);
      }
      $loglike_new += $fsum;

      # expected transitions:
      ($trans0[$p], $trans[$p]) = count_transitions(\@parent, $timepoint[$p], \@compatibles, \%theta, \@forward, \@backward, $clones[$p], $eps_p, $eps_n, $fsum);

      # expected flips:
      for $t (0 .. $#{$timepoint[$p]}) {
        for $c (0 .. $#{$clones[$p][$t]}) {
          for $v (1 .. $#parent) {
            for $s (0 .. $#compatibles) {
              $a = $compatibles[$s][$v];
              $b = $clones[$p][$t][$c][$v];
              $flips[$v][$a][$b] += exp($forward[$t][$s] + $backward[$t][$s] - $fsum);
                 
            }
          }
        }
      }

    } # end for $p


    # M step:
    # -------

    @lambda = MLE_lambda(\@lambda, \@mu, \@offset, \@parent, \@trans0, \@trans, \@timepoint);
    ($eps_p, $eps_n) = MLE_eps(\@flips, scalar(@parent), $uniform_eps_p, $uniform_eps_n); 

    print "\nlambda = ", join(', ', @lambda[1..$#parent]);
    print "\neps pos = ", join(', ', @{$eps_p}[1..$#parent]);
    print "\neps neg = ", join(', ', @{$eps_n}[1..$#parent]);
    
    print "\n\nlog-likelihood = $loglike_new\n\n";

  }


  # Viterbi path:
  # -------------

  my ($logprob, @states);
  if ($do_viterbi) {
    for $p (0 .. $#clones) {
      @theta0 = initial_state_loglike(\@compatibles, \@parent, \@lambda, $mu[$p], $offset[$p], $timepoint[$p]);
      %theta = all_transition_matrices(\@timepoint, \@parent, \@lambda, $mu[$p], $offset[$p], $compatibles, \@is_pred_state);
     ($logprob, @{$states[$p]}) = viterbi(\@theta0, \%theta, \@compatibles, $timepoint[$p], $eps_p, $eps_n, $clones[$p]);
    }
  }

  return($loglike_new, \@lambda, \@mu, \@offset, $eps_p, $eps_n, \@states);
}



sub count_transitions {

  # expected transition counts at all timepoints  t --> t+1

  my @parent = @{shift @_};
  my @timepoint = @{shift @_};
  my @compatibles = @{shift @_};
  my %theta = %{shift @_};
  my @forward = @{shift @_};
  my @backward = @{shift @_}; 
  my @clones = @{shift @_};
  my $eps_p = shift;
  my $eps_n = shift;
  my $dataloglike = shift;

  my ($t, $v, $r, $s, $delta_t, $a, $b);
  my @trans0;

  # expected counts of initial states (t == 0)
  for $s (0 .. $#compatibles) {
    for $v (1 .. $#parent) {  # consider edge  v <-- pa(v)
      $a = $compatibles[$s][$v];           # i_v
      $b = $compatibles[$s][$parent[$v]];  # i_pa(v)
      $trans0[$v][$a][$b] += exp($forward[0][$s] + $backward[0][$s] - $dataloglike);
    }
  }

  # expected transitions t --> t+1,  t >= 0
  my @trans;
  for $t (0 .. $#timepoint-1) {
    for $v (1 .. $#parent) {
      $trans[$t][$v][0][0] = $trans[$t][$v][0][1] = 0.0;
      $trans[$t][$v][1][0] = $trans[$t][$v][1][1] = 0.0;
      for $r (0 .. $#compatibles) {
        for $s (0 .. $#compatibles) {
          if ($compatibles[$s][$parent[$v]] == 1) {
            $delta_t = $timepoint[$t+1] - $timepoint[$t];
            if (defined($theta{$delta_t}->[$r][$s])) {
              $a = $compatibles[$r][$v];
              $b = $compatibles[$s][$v];
              $trans[$t][$v][$a][$b] += exp($forward[$t][$r] + $theta{$delta_t}->[$r][$s] + emission($compatibles[$s], $clones[$t+1], $eps_p, $eps_n) + $backward[$t+1][$s] - $dataloglike);
            }
          }
        }
      }
    }
  }

  return (\@trans0, \@trans);
}



sub MLE_eps {

  my @flips = @{shift @_};
  my $N_v = shift;
  my $uniform_eps_p = shift;
  my $uniform_eps_n = shift;

  my (@eps_p, @eps_n, $v);

  my @allflips = ();
  for $v (1 .. $N_v-1) {  # summarize over events
    $eps_n[$v] = min(1 - $EpsEps, $flips[$v][1][0] / ($flips[$v][1][0] + $flips[$v][1][1]));
    $eps_n[$v] = max($EpsEps, $eps_n[$v]);

    $eps_p[$v] = max($EpsEps, $flips[$v][0][1] / ($flips[$v][0][0] + $flips[$v][0][1]));
    $eps_p[$v] = min(1 - $EpsEps, $eps_p[$v]);

    $allflips[0][0] += $flips[$v][0][0];
    $allflips[0][1] += $flips[$v][0][1];
    $allflips[1][0] += $flips[$v][1][0];
    $allflips[1][1] += $flips[$v][1][1];
  }

  if ($uniform_eps_p) {
    for $v (1 .. $N_v-1) {  
      $eps_p[$v] = $allflips[0][1] / ($allflips[0][1] + $allflips[0][0]);
    }
  }

  if ($uniform_eps_n) {
    for $v (1 .. $N_v-1) {  
      $eps_n[$v] = $allflips[1][0] / ($allflips[1][0] + $allflips[1][1]);
    }
  }

  return (\@eps_p, \@eps_n);
}



sub exploglike_of_mu {

  # expected log-likelihood for fixed lambda
  # as a function of $mu[p]

  my $mu = shift;  # for fixed patient

  my @lambda      = @{$FIXEDPARAM[0]};
  my @parent      = @{$FIXEDPARAM[1]};
  my @trans0      = @{$FIXEDPARAM[2]};
  my @trans       = @{$FIXEDPARAM[3]};
  my @timepoint   = @{$FIXEDPARAM[4]};

  my $N_t = scalar(@timepoint);
  my $N_v = scalar(@parent) - 1;
  my $exploglike = 0.0;
  my ($t, $v, $delta_t);

  if ($timepoint[0] > 0) {  # first timepoint after 0?
    $delta_t = $timepoint[0];
    for $v (1 .. $N_v) {
      $exploglike += -$lambda[$v]*$mu*$delta_t*$trans0[$v][0][1] + log(1 - exp(-$lambda[$v]*$mu*$delta_t))*$trans0[$v][1][1];
    } 
  }

  for $t (0 .. $N_t-2) {
    $delta_t = $timepoint[$t+1] - $timepoint[$t];
    for $v (1 .. $N_v) {
      $exploglike += -$lambda[$v]*$mu*$delta_t*$trans[$t][$v][0][0] + log(1 - exp(-$lambda[$v]*$mu*$delta_t))*$trans[$t][$v][0][1];
    }
  }

  return -$exploglike;
}



sub exploglike_of_lambda {

  # expected log-likelihood for fixed @mu
  # as a function of lambda[]

  my @lambda = (-1, @_);

  my @mu        = @{$FIXEDPARAM[0]};
  my @offset    = @{$FIXEDPARAM[1]};
  my @parent    = @{$FIXEDPARAM[2]};
  my @trans0    = @{$FIXEDPARAM[3]};
  my @trans     = @{$FIXEDPARAM[4]};
  my @timepoint = @{$FIXEDPARAM[5]};

  for my $v (1 .. $#parent) {
    $lambda[$v] = exp($lambda[$v]);
  }

  my ($v, $t, $p, $delta_t);
  my $exploglike = 0.0;

  for $p (0 .. $#timepoint) { 

    if ($timepoint[$p][0] > 0) {  # first timepoint after 0?
      $delta_t = $timepoint[$p][0];
      for $v (1 .. $#parent) {
        $exploglike += -$lambda[$v]*$mu[$p]*$delta_t*$trans0[$p][$v][0][1] + log(1 - exp(-$lambda[$v]*$mu[$p]*$delta_t))*$trans0[$p][$v][1][1];
      } 
    }

    for $t (0 .. $#{$timepoint[$p]}-1) {
      $delta_t = $timepoint[$p][$t+1] - $timepoint[$p][$t];
      for $v (1 .. $#parent) {
        $exploglike += -$lambda[$v]*$mu[$p]*$delta_t*$trans[$p][$t][$v][0][0] + log(1 - exp(-$lambda[$v]*$mu[$p]*$delta_t))*$trans[$p][$t][$v][0][1];
      }
    }

  }

  return -$exploglike;
}



sub MLE_lambda {

  # ML estimation of lambda[] 
  # via Nelder and Mead's downhill simplex  method

  my @lambda = @{shift @_};
  @FIXEDPARAM = @_;

  my (@guess, @scale);
  for my $v (1 .. $#lambda) {
    $guess[$v-1] = log($lambda[$v]);
    $scale[$v-1] = 1.0;
  }

  my ($argmin, $y) = MinimiseND(\@guess, \@scale, \&exploglike_of_lambda, 1e-5, 100);
  
  for my $v (1 .. $#lambda) {
    $lambda[$v] = exp($argmin->[$v-1]);   
  }

  return @lambda;
}



sub viterbi {

  # For a fixed patient, compute the most probable state path
  #   viterbi_path[t][s] == argmax_states Prob(clones[], states[t][s])
  # and its log probability
  #   viterbi_loglike[t][s] == max_states log Prob(clones[], states[t][s])

  my @theta0 = @{shift @_};
  my %theta  = %{shift @_};
  my @compatibles = @{shift @_};
  my @timepoint = @{shift @_};  # patient fixed
  my $eps_p = shift;
  my $eps_n = shift;
  my @clones = @{shift @_};  # patient fixed

  
  my ($t, $s, $s_prev);
  my (@viterbi, @pointer);  

  # initialization at t=0:
  for $s (0 .. $#compatibles) {
    $viterbi[0][$s] += $theta0[$s] + emission($compatibles[$s], $clones[0], $eps_p, $eps_n);
  }

  # forward recursion over t:
  for $t (1 .. $#timepoint) {
    my $delta_t = $timepoint[$t] - $timepoint[$t-1];
    my @v = ();
    for $s (0 .. $#compatibles) {
      # consider all transitions from all possible states at time t-1
      for $s_prev (0 .. $#compatibles) {
        if (defined($theta{$delta_t}->[$s_prev][$s])) {
          $v[$s_prev] = $viterbi[$t-1][$s_prev] + $theta{$delta_t}->[$s_prev][$s];
        }
        else {
          $v[$s_prev] = undef;
        }
      }
      ($viterbi[$t][$s], $pointer[$t][$s]) = max(@v);
      $viterbi[$t][$s] += emission($compatibles[$s], $clones[$t], $eps_p, $eps_n);  # add emission likelihood of clones
    }
  }

  # termination:
  my $viterbi_logprob;
  my @viterbi_path;
  my @v = ();
  for $s (0 .. $#compatibles) {
    $v[$s] = $viterbi[$#timepoint][$s];
  }
  ($viterbi_logprob, $viterbi_path[$#timepoint]) = max(@v);

  # traceback:
  for ($t=$#timepoint; $t>=1; $t--) {
    $viterbi_path[$t-1] = $pointer[$t][$viterbi_path[$t]];
  }

  return ($viterbi_logprob, @viterbi_path);
}



sub forward_pass {

  # For a fixed patient, compute
  # forward[t][s] == log Prob(clones[0...t], states[t]==s)

  my @theta0 = @{shift @_};
  my %theta  = %{shift @_};
  my @compatibles = @{shift @_};
  my @timepoint = @{shift @_};  # patient fixed
  my $eps_p = shift;
  my $eps_n = shift;
  my @clones = @{shift @_};  # patient fixed

  
  my ($t, $s, $s_prev);
  my @forward;
  
  # initialization at t=0:
  for $s (0 .. $#compatibles) {
    $forward[0][$s] += $theta0[$s] + emission($compatibles[$s], $clones[0], $eps_p, $eps_n);
  }

  # forward recursion over t:
  for $t (1 .. $#timepoint) {
    my $delta_t = $timepoint[$t] - $timepoint[$t-1];

    for $s (0 .. $#compatibles) {
      # consider all transitions from all possible states at time t-1
      $forward[$t][$s] += $forward[$t-1][0] + $theta{$delta_t}->[0][$s];  # make sure any state is reachable from the first one!

      for $s_prev (1 .. $#compatibles) {
        if (defined($theta{$delta_t}->[$s_prev][$s])) {
          $forward[$t][$s] = logsum($forward[$t][$s], $forward[$t-1][$s_prev] + $theta{$delta_t}->[$s_prev][$s]);
        }
      }
      $forward[$t][$s] += emission($compatibles[$s], $clones[$t], $eps_p, $eps_n);  # add emission likelihood of clones
    }
  }

  return @forward;
}




sub backward_pass {

  # For a fixed patient, compute
  # backward[t][s] == log Prob(clones[t+1...N_t-1] | states[t]==s)

  my @theta0 = @{shift @_};
  my %theta = %{shift @_};
  my @compatibles = @{shift @_};
  my @timepoint = @{shift @_};  # patient fixed
  my $eps_p = shift;
  my $eps_n = shift;
  my @clones = @{shift @_};  # patient fixed

  
  my ($t, $s, $s_next);
  my @backward;

  # initialization at t=N_t:
  for $s (0 .. $#compatibles) {
    $backward[$#timepoint][$s] += 0.0;
  }

  # backward recursion over t:
  for ($t=$#timepoint-1; $t>=0; $t--) {
    my $delta_t = $timepoint[$t+1] - $timepoint[$t];

    for $s (0 .. $#compatibles) {
      # consider all transitions to all possible states at time t+1
      $backward[$t][$s] += $backward[$t+1][$#compatibles] + $theta{$delta_t}->[$s][$#compatibles] + emission($compatibles[$#compatibles], $clones[$t+1], $eps_p, $eps_n);  # make sure the last state is always reachable!

      for $s_next (0 .. $#compatibles-1) {
        if (defined($theta{$delta_t}->[$s][$s_next])) {
          $backward[$t][$s] = logsum($backward[$t][$s], $backward[$t+1][$s_next] + $theta{$delta_t}->[$s][$s_next] + emission($compatibles[$s_next], $clones[$t+1], $eps_p, $eps_n));
        }
      }
    }
  }

  return @backward;
}



sub initial_state_loglike {

  # initial log-likelihood of states at time t==0.

  my @compatibles = @{shift @_};
  my $parent = shift;
  my $lambda = shift;
  my $mu = shift;
  my $offset = shift;
  my @timepoint = @{shift @_};

  my $N_c = scalar(@compatibles);
  my @theta0;

  # uniform:
  #my $loglike = -log($N_c);
  #for my $s (0 .. $N_c-1) {
  #  $theta0[$s] = $loglike;  # uniform
  #}

  # centered on 0...0 state:
  #$theta0[0] = log(1 - (($N_c-1)*1e-100));
  #for my $s (1 .. $N_c-1) {
  #  $theta0[$s] = log(1e-100);
  #}

  # according to tree model with 0...0 at time 0:
  for my $s (0 .. $#compatibles) {
    $theta0[$s] = loglike_pattern($compatibles[$s], $parent, $lambda, $mu, $offset, $timepoint[0]);
  }

  return @theta0;
}


sub loglike_pattern {

  my @pattern = @{shift @_}; 
  my @parent  = @{shift @_};
  my @lambda  = @{shift @_};
  my $mu      = shift;
  my $offset  = shift;
  my $delta_t = shift;

  my $loglike = 0.0;

  for my $v (1 .. $#parent) {
    if ($pattern[$parent[$v]] == 1) {  # 1 -->
      if ($pattern[$v] == 1) {  # 1 --> 1
        if ($delta_t > 0) {  # in positive time
          $loglike += log(1 - exp(-$lambda[$v]*$delta_t));
        }
        else {  # in zero time
          return -100000;
        }
      }
      else {  # 1 --> 0
        $loglike += -$lambda[$v]*$delta_t;
      }
    }
  }

  return $loglike;
}



sub all_transition_matrices {

  my @timepoint     = @{shift @_};
  my $parent        = shift;
  my $lambda        = shift;
  my $mu            = shift;
  my $offset        = shift;
  my $compatibles   = shift;
  my $is_pred_state = shift;

  # time intervals:
  my %theta;
  for my $p (0 .. $#timepoint) {
    for my $t (1 .. $#{$timepoint[$p]}) {
      $theta{$timepoint[$p][$t] - $timepoint[$p][$t-1]} = 1;
    }
  }

  for my $delta_t (keys %theta) { 
    my @theta = transition_matrix($delta_t, $parent, $lambda, $mu, $offset, $compatibles, $is_pred_state);
    $theta{$delta_t} = \@theta;
  }

  return %theta;
}




sub transition_matrix {

  # (log) transition matrix for a given time interval delta_t

  my $delta_t       = shift;
  my @parent        = @{shift @_};
  my @lambda        = @{shift @_};
  my $mu            = shift;
  my $offset        = shift;
  my @compatibles   = @{shift @_};
  my @is_pred_state = @{shift @_};

  my ($s1, $s2, $v, @pattern1, @pattern2);
  my @theta;

  for $s1 (0 .. $#compatibles) {
    for $s2 (0 .. $#compatibles) {
      if ($is_pred_state[$s1][$s2]) {  # s1 <= s2 ?      
        @pattern1 = @{$compatibles[$s1]};
        @pattern2 = @{$compatibles[$s2]};

        $theta[$s1][$s2] = 0.0;
        for $v (1 .. $#parent) {
          if ($pattern2[$parent[$v]] == 1) {  # predecessor mutation is present
            if ($pattern2[$v] == 1) {  # mutation has occurred
              if ($pattern1[$v] == 0) { # mutation has not been present before
                $theta[$s1][$s2] += log(1 - exp(-($lambda[$v] * $mu)*$delta_t));
              }
            }
            else {  # mutation has not occurred
              $theta[$s1][$s2] += -($lambda[$v] * $mu)*$delta_t;
            }
          }
        }

      }
      else {  # transition impossible
        undef($theta[$s1][$s2]);
      }
  
    }
  }

  return @theta;
}



sub emission {

  # log-likelihood of a set of clones 
  # for one patient at one time point
  # given the state and the model parameters

  my @states = @{shift @_};  # pattern of state
  my @clones = @{shift @_};  # emitted clones
  my @eps_p = @{shift @_};
  my @eps_n = @{shift @_};

  my ($clone, @pattern, $v);

  # count flips:  
  my $loglike = 0.0;
  for $clone (@clones) {
    my @pattern = @{$clone};
    for $v (1 .. $#states) {
      if ($states[$v] == 0) {
        if ($pattern[$v] == 0) {
          $loglike += log(1 - $eps_p[$v]);
        }
        else {
          $loglike += log($eps_p[$v]);
        }
      }
      else {  # state == 1
        if ($pattern[$v] == 1) {
          $loglike += log(1 - $eps_n[$v]);
        }
        else {
          $loglike += log($eps_n[$v]);
        }
      }   
    }
  }

  return $loglike;
}


sub guess_mu {

  my $N_p = shift;

  my @mu;
  for my $p (0 .. $N_p-1) {
    $mu[$p] = 1.0;
  }

  return @mu;
}


sub guess_offset {

  my $N_p = shift;

  my @offset;
  for my $p (0 .. $N_p-1) {
    $offset[$p] = 0.0;
  }

  return @offset;
}



sub guess_eps {

  my $N_v = shift;

  my (@eps_p, @eps_n);

  $eps_p[0] = $eps_n[0] = -1;
  for my $v (1 .. $N_v) {
    $eps_p[$v] = $eps_n[$v] = $InitialEps;
  }    

  return (\@eps_p, \@eps_n);
}



sub load_model {

  my $filestem = shift;
  my $sampling_mode = shift;
  my $sampling_param = shift;
  
  my ($events, $parent, $edgeprob) = load_mtree($filestem);
  my @lambda = waiting_times($sampling_mode, $sampling_param, $edgeprob);
  my ($compatibles, $is_pred_state) = load_compatibles($filestem, $parent);

  return ($events, $parent, \@lambda, $compatibles, $is_pred_state);
}



sub load_data {

  my $filestem = shift;

  my @patient = load_patient($filestem);
  my @time    = load_week($filestem);
  my @pattern = load_pattern($filestem);


  if ($option{R}) {
    my $from = 2300;
    my $to   = 2700;
    @patient = @patient[$from..$to];
    @time    = @time[$from..$to];
    @pattern = @pattern[$from..$to];
  }


  # build data structures:
  my ($p, $t, $i);

  # all patients:
  my %ind;
  for (@patient) {  # remove duplicates
    $ind{$_} = 1;
  }
  my @allpatients = sort {$a cmp $b} keys %ind;  # sort
  my %index_of_patient;
  for $p (0 .. $#allpatients) {  # build reverse index
    $index_of_patient{$allpatients[$p]} = $p;
  }
  print "\nPatients = {", join(", ", @allpatients), "}\n";


  # all time points per patient:
  my @timepoint;
  for $i (0 .. $#patient) {  # collect per patient
    push @{$timepoint[$index_of_patient{$patient[$i]}]}, $time[$i];
  }

  my @index_of_timepoint;

  for $p (0 .. $#allpatients) {
    %ind = ();
    for (@{$timepoint[$p]}) {  # remove duplicates
      $ind{$_} = 1;
    }
    @{$timepoint[$p]} = sort {$a <=> $b} keys %ind;  # sort

    for $t (0 .. $#{$timepoint[$p]}) {  # biuld reverse index
      $index_of_timepoint[$p]->{$timepoint[$p][$t]} = $t;
    }
  }


  # all clones per patient and time point:
  my @clones;
  for $i (0 .. $#patient) {
    $p = $index_of_patient{$patient[$i]};
    $t = $index_of_timepoint[$p]->{$time[$i]};
    push @{$clones[$p][$t]}, $pattern[$i];
  }


  return (\@allpatients, \@timepoint, \@clones);
}



sub load_pattern {

  my $filestem = shift;

  my @X;

  open PATTERN, "$filestem.pat"
    or die "Can't open pattern file $filestem.pat!\n";
  <PATTERN>;  # skip dimensions
  while (<PATTERN>) {
    my @pat = trim(split /\s+/, $_);
    push @X, \@pat; 
  }
  close PATTERN;

  return @X;
}



sub load_mtree {

  my $filestem = shift;

  my ($i, $j);
  my @parent;
  my @theta;


  # load model file
  open MODEL, "$filestem.model"
    or die "Can't open model file $filestem.model!\n";
  my ($K, @weights) = split /\s+/, <MODEL>;
  if (($K < 1) || ($K > 2)) {
    die "$filestem.model has to be a single tree model (possibly mixed with a noise component)!\n";
  }
  my @model = <MODEL>;
  close MODEL;

  # extract adjacency matrix
  my @adj;  # adjacency matrix
  my ($n, $dummy) = split /\s+/, $model[0];
  for $i (($K-1)*$n+$K .. $K*$n+$K-1) { 
    my @prob = trim(split /\s+/, $model[$i]);
    shift @prob;  # remove initial white space
    push @adj, \@prob;
  }

  # build parent vector 
  # and parameter vector theta (conditional probabilities)
  $parent[0] = -1;
  $theta[0] = -1;
  for $i (0 .. $#adj) {
    for $j (0 .. $#{$adj[0]}) {
      if ($adj[$i][$j] > 0) {
        #print "$i --> $j\n";
        $parent[$j] = $i;
        $theta[$j]  = $adj[$i][$j];
      }
    }
  }

  # load events
  my @events;
  if (-e "$filestem.prf") {
    open PRF, "$filestem.prf"
      or die "Can't open profile file $filestem.prf!";
    @events = trim(<PRF>);
    close PRF;
  }
  else {
    @events = (0 .. $#parent);
  }


  return (\@events, \@parent, \@theta);
}



sub load_compatibles {

    my $filestem = shift;
    my @parent = @{shift @_};

    my $n = scalar(@parent) - 1;
    my ($j, @a);
    for $j (0 .. $n) {
        $a[$j] = 0;
    }
    
    my @compatibles;
    do {
        # check n-tuple for compatibility:
	my $is_compatible = 1;
	my $save_a0 = $a[0];
	$a[0] = 1;
	for my $i (1 .. $n) {
	    $a[0] = 1;
	    if (($a[$i] == 1) && ($a[$parent[$i]] == 0)) {
		$is_compatible = 0;
		last;
	    }
	}
	push @compatibles, [ @a ]  if $is_compatible;	
	$a[0] = $save_a0;

        $j = $n;
        while (($a[$j] == 1)) {  # carry:
            $a[$j] = 0;
            $j--;
        }

        $a[$j]++;  # increase binary number $a
	
    } until ($j == 0);
    

    # determine valid transitions (i.e. those respecting the lattice structure)
    my @is_pred_state;  # is predecessor state ("<=") in lattice
    for my $r (0 .. $#compatibles) {
	for my $s (0 .. $#compatibles) {
	    $is_pred_state[$r][$s] = 1;
	    my $v = 1;
	    while ($v <= $#parent && $is_pred_state[$r][$s] == 1) {
		if (($compatibles[$r][$v] == 1) && ($compatibles[$s][$v] == 0)) {
		    $is_pred_state[$r][$s] = 0;
		}
		$v++;
	    }
	}
    }
    
    return (\@compatibles, \@is_pred_state);
}



sub load_patient {

  my $filestem = shift;

  open PATIENT, "$filestem.patient"
    or die "Can't open patient file $filestem.patient!\n";
  my @patient = trim(<PATIENT>);
  close PATTERN;

  return @patient;
}



sub load_week {

  my $filestem = shift;

  open WEEK, "$filestem.week"
    or die "Can't open week file $filestem.week!\n";
  my @week = trim(<WEEK>);
  close WEEK;

  return @week;
}



sub print_model {

  my @parent = @{shift @_};
  my @lambda = @{shift @_};
  my @compatibles = @{shift @_};
  my @is_pred_state = @{shift @_};

  my $i;

  print "\nParent vector = (", join(", ", @parent[1..$#parent]), ")\n";
  print "Lambda = ", join(", ", @lambda[1..$#parent]), "\n";
  print "Sampling mode = $sampling_mode,  parameter = $sampling_param\n\n";

  print "Compatible states =\n";
  for my $i (0 .. $#parent) {
    printf "%2d", $i;
  }
  print " ";
  for $i (0 .. $#{$compatibles}) {
    if ($i % 5 == 0) {
      printf "%2d", $i;
    }
    else {
      print "  ";
    }
  } 
  print "\n";

  my $count = 0;
  for $i (0 .. $#{$compatibles}) {
    print " ", join(" ", @{$compatibles[$i]}), "  ";
    for my $j (0 .. $#{$compatibles}) {
      if ($is_pred_state[$i][$j] > 0) {
        print "* ";
        $count++;
      }
      else {
        print "  ";
      }
    }
    printf "\n";
  }
  print scalar(@{$compatibles}), " compatible states, $count transitions.\n\n";

}



sub print_param {

  my $b           = shift;
  my $filestem    = shift;
  my @allpatients = @{shift @_};
  my @events      = @{shift @_};
  my $loglike     = shift;
  my @lambda      = @{shift @_};
  my @mu          = @{shift @_};
  my %mu_list     = %{shift @_};
  my @offset      = @{shift @_};
  my %offset_list = %{shift @_};
  my @eps_p       = @{shift @_};
  my @eps_n       = @{shift @_};

  my $N_v = scalar(@events);
  my ($p, $patient);

  if ($b == 0) {  # create files, print headers
    open LL, ">$filestem.l"
      or die "Can't open output file $filestem.l!";
    close LL;

    open LA, ">$filestem.lambda"
      or die "Can't open output file $filestem.lambda!";
    print LA join("\t", @events[1..$N_v-1]), "\n";
    close LA;

    open EPSP, ">$filestem.epsp"
      or die "Can't open output file $filestem.epsp";
    print EPSP join("\t", @events[1..$N_v-1]), "\n";
    close EPSP;

    open EPSN, ">$filestem.epsn"
      or die "Can't open output file $filestem.epsn";
    print EPSN join("\t", @events[1..$#events]), "\n";
    close EPSN;
  }


  # append rows:

  open LA, ">>$filestem.lambda"
    or die "Can't open output file $filestem.lambda!";
  print LA join("\t", @lambda[1..$N_v-1]), "\n";
  close LA;

  open LL, ">>$filestem.l"
    or die "Can't open output file $filestem.l!";
  print LL $loglike, "\n";
  close LL;

  open EPSP, ">>$filestem.epsp"
    or die "Can't open output file $filestem.epsp";
  print EPSP join("\t", @eps_p[1..$N_v-1]), "\n";
  close EPSP;

  open EPSN, ">>$filestem.epsn"
    or die "Can't open output file $filestem.epsn";
  print EPSN join("\t", @eps_n[1..$N_v-1]), "\n";
   close EPSN;

}



sub print_viterbi {

  my $filestem = shift;
  my @allpatients = @{shift @_};
  my @timepoint = @{shift @_};
  my $N_v = shift;
  my @mu = @{shift @_};
  my @offset = @{shift @_};
  my @states = @{shift @_};
  my @clones = @{shift @_}; 

  open VIT, ">$filestem.viterbi"
    or die "Can't open output file $filestem.viterbi!\n";
  print VIT "Patient\tWeek\tPopSt\tClone\n";
  for my $p (0 .. $#allpatients) {
    for my $t (0 .. $#{$timepoint[$p]}) {
      printf VIT "$allpatients[$p]\t$timepoint[$p][$t]\t";
      print VIT "$states[$p][$t]\t";
      print VIT join('', @{$compatibles->[$states[$p][$t]]}[1..$N_v-1]), "\n";

      # include data in output:
      for (@{$clones[$p][$t]}) {
	  printf VIT "$allpatients[$p]\t$timepoint[$p][$t]\t\t";
      	  print VIT join('', @{$_}[1..$N_v-1]), "\n";
      }
    }
  }
  close VIT;

}



sub waiting_times {

  # Ccompute waiting times (parameter lambda
  # of the exponential distribution) on each edge

  my $sampling_mode  = shift;
  my $sampling_param = shift;
  my @theta          = @{shift @_};

  my @lambda = (-1);;

  if ($sampling_mode eq 'CONSTANT') { 
    # sampling_param == t_0
    # lambda = -log(1 - p) / t_0
    for my $i (1 .. $#theta) {
      $lambda[$i]= -log(1 - $theta[$i]) / $sampling_param;
    }
  }

  if ($sampling_mode eq 'EXPONENTIAL') {
    # sampling_param == lambda_s
    # lambda = (lambda_s * p) / (1 - p)
    for my $i (1 .. $#theta) {
      $lambda[$i] = ($sampling_param * $theta[$i]) / (1 - $theta[$i]);
    }
  }
    
  return @lambda;
}



sub resample {

  my @allpatients = @{shift @_};
  my @timepoint   = @{shift @_};
  my @clones      = @{shift @_};

  my (@allpatients_b, @timepoint_b, @clones_b);

  my $N_p = scalar(@allpatients);
  for my $p (0 .. $N_p-1) {
    my $p_rand =  intrand(0, $N_p - 1);
    push @allpatients_b, $allpatients[$p_rand];
    push @timepoint_b, $timepoint[$p_rand];
    for my $t (0 .. $#{$timepoint[$p_rand]}) {
      my $N_c = scalar(@{$clones[$p_rand][$t]});
      for my $c (0 .. $N_c-1) {
        my $c_rand = intrand(0, $N_c - 1);
        push @{$clones_b[$p][$t]}, \@{$clones[$p_rand][$t][$c_rand]};
      }
    }
  }

  return (\@allpatients_b, \@timepoint_b, \@clones_b);
}



sub logsum {

  # computes log(x + y) from (log x) and (log y)

  my $lx = shift;  # log x
  my $ly = shift;  # log y

  #return log(exp($lx) + exp($ly));  # naive

  return ($lx >= $ly) ? ($lx + log(1 + exp($ly - $lx))) : ($ly + log(1 + exp($lx - $ly)));
}



sub intrand {

  # random integer number between and including $a and $b

  my $a = shift;
  my $b = shift;

  return $a + int(rand($b-$a+1));
}



sub trim {  # remove white spaces to the left and right

  my @input = @_;

  for (@input) {
    s/^\s+//;
    s/\s+$//;
  }

  return wantarray ? @input : $input[0];
}



sub max {

  # Maximum of a vector containing undef's

  my @input = @_;

  my ($max, $argmax);

  for my $i (0 .. $#input) {
    if (defined($input[$i])) {
      if (! defined($argmax)) {
        $max = $input[$i];
        $argmax = $i;
      }
      elsif ($input[$i] > $max) {  # same action, but avoids warning
        $max = $input[$i];
        $argmax = $i;
      }
    }
  }

  return wantarray ? ($max, $argmax) : $max;
}



sub min {

  # Minimum of a vector containing undef's

  my @input = @_;

  my ($min, $argmin);

  for my $i (0 .. $#input) {
    if (defined($input[$i])) {
      if (! defined($argmin)) {
        $min = $input[$i];
        $argmin = $i;
      }
      elsif ($input[$i] < $min) {  # same action, but avoids warning
        $min = $input[$i];
        $argmin = $i;
      }
    }
  }

  return wantarray ? ($min, $argmin) : $min;
}

